import logging

from pystorz.internal import constants
from pystorz.store import store

log = logging.getLogger(__name__)

from pystorz.store import options


class InternalStore(store.Store):
    def __init__(self, schema: store.SchemaHolder, data: store.Store):
        self.Schema = schema
        self.Store = data

    def Create(self, obj: store.Object, *opt: options.CreateOption) -> store.Object:
        if obj is None:
            raise Exception(constants.ErrObjectNil)

        log.info("create {} {}".format(
            obj.Metadata().Kind(),
            obj.PrimaryKey()))

        original = self.Schema.ObjectForKind(obj.Metadata().Kind())
        if original is None:
            raise Exception(constants.ErrInvalidPath)

        if isinstance(obj, store.ExternalHolder):
            original.ExternalInternalSet(obj.External())

        return self.Store.Create(original, *opt)

    def Update(
        self,
        identity: store.ObjectIdentity,
        obj: store.Object,
        *opt: options.UpdateOption,
    ) -> store.Object:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        if obj is None:
            raise Exception(constants.ErrObjectNil)

        log.info("update {}".format(identity.Path()))

        original = self.Store.Get(identity)
        if isinstance(obj, store.ExternalHolder):
            original.ExternalInternalSet(obj.External())

        if original.Metadata().Kind() != obj.Metadata().Kind():
            raise Exception(constants.ErrObjectIdentityMismatch)

        return self.Store.Update(identity, original, *opt)

    def Delete(self, identity: store.ObjectIdentity, *opt: options.DeleteOption):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info("delete {}".format(identity.Path()))
        return self.Store.Delete(identity, *opt)

    def Get(
        self, identity: store.ObjectIdentity, *opt: options.GetOption
    ) -> store.Object:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info("get {}".format(identity.Path()))
        return self.Store.Get(identity, *opt)

    def List(
        self, identity: store.ObjectIdentity, *opt: options.ListOption
    ) -> store.ObjectList:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        log.info("list {}".format(identity.Path()))
        return self.Store.List(identity, *opt)



import json
from datetime import datetime
from pystorz.store import store, utils


class Meta:
    def Kind(self) -> str:
        pass

    def Identity(self) -> store.ObjectIdentity:
        pass

    def Created(self) -> datetime:
        pass

    def Updated(self) -> datetime:
        pass

    def Revision(self) -> int:
        pass

    def ToJson(self) -> str:
        pass

    def FromDict(self, d: dict) -> None:
        pass

    def ToDict(self) -> str:
        pass


class MetaSetter:
    def SetKind(self, kind: str) -> None:
        pass

    def SetIdentity(self, identity: store.ObjectIdentity) -> None:
        pass

    def SetCreated(self, created: datetime) -> None:
        pass

    def SetUpdated(self, updated: datetime) -> None:
        pass

    def SetRevision(self, revision: int) -> None:
        pass


# class MetaHolder:
#     def metadata(self) -> Meta:
#         pass


class metaWrapper(Meta, MetaSetter):
    def __init__(self, kind):
        self.revision_ = 0
        self.kind_ = kind
        self.identity_ = store.ObjectIdentityFactory()
        self.created_ = ""
        self.updated_ = ""

    def Kind(self) -> str:
        return self.kind_

    def Created(self) -> datetime:
        return utils.datetime_parse(self.created_)

    def Updated(self) -> datetime:
        return utils.datetime_parse(self.updated_)

    def Identity(self) -> store.ObjectIdentity:
        return self.identity_

    def Revision(self) -> int:
        return self.revision_

    def SetKind(self, kind: str) -> None:
        self.kind_ = kind

    def SetIdentity(self, identity: store.ObjectIdentity) -> None:
        self.identity_ = identity

    def SetCreated(self, created: datetime) -> None:
        self.created_ = utils.datetime_string(created)

    def SetUpdated(self, updated: datetime) -> None:
        self.updated_ = utils.datetime_string(updated)

    def SetRevision(self, revision: int) -> None:
        self.revision_ = revision

    def ToJson(self) -> str:
        return json.dumps(self.ToDict())

    def ToDict(self) -> str:
        return {
            "kind": self.Kind(),
            "identity": str(self.Identity()),
            "created": self.created_,
            "updated": self.updated_,
            "revision": self.revision_,
        }

    def FromDict(self, d: dict) -> None:
        self.SetKind(d["kind"])
        self.SetIdentity(store.ObjectIdentityFactory())
        self.Identity().FromString(d["identity"])
        self.created_ = d["created"]
        self.updated_ = d["updated"]
        self.revision_ = d["revision"]


def MetaFactory(kind: str) -> Meta:
    return metaWrapper(kind)
    


import logging
import threading
import sqlparse

from pystorz.internal import constants
from pystorz.store import store, options, utils

log = logging.getLogger(__name__)


class SqlStore(store.Store):
    def __init__(self, Schema, connector):
        self._schema = Schema
        self._makeConnection = connector
        self._connection_cache = {}

    def _connection(self):
        tid = threading.get_ident()
        if tid not in self._connection_cache:
            self._connection_cache[tid] = self._makeConnection()
            self._prepareTables(self._connection_cache[tid])

        return self._connection_cache[tid]

    def Create(self, obj, *opt):
        if obj is None:
            raise Exception(constants.ErrObjectNil)

        log.info("create {} {}".format(
            obj.Metadata().Kind(),
            obj.PrimaryKey()))

        # copt = options.CommonOptionHolderFactory()
        # for o in opt:
        #     o.ApplyFunction()(copt)

        lk = obj.Metadata().Kind().lower()
        path = "{}/{}".format(lk, obj.PrimaryKey())
        existing = None
        try:
            existing = self.Get(store.ObjectIdentity(path))
        except Exception as e:
            pass

        if existing is not None:
            raise Exception(constants.ErrObjectExists)

        db = self._connection()

        try:
            # Start a transaction
            db.execute("BEGIN")
            cursor = db.cursor()

            self._setIdentity(
                cursor,
                obj.Metadata().Identity().Path(),
                obj.PrimaryKey(),
                obj.Metadata().Kind(),
            )

            self._setObject(cursor, obj.PrimaryKey(), obj.Metadata().Kind(), obj)

            # Commit the transaction
            db.commit()
            return obj.Clone()
        except Exception as e:
            db.rollback()
            raise e

    def Update(self, identity, obj, *opt):
        # copt = options.CommonOptionHolderFactory()
        # for o in opt:
        #     o.ApplyFunction()(copt)

        if identity is None:
            raise Exception(constants.ErrInvalidPath)

        if obj is None:
            raise Exception(constants.ErrObjectNil)

        log.info("update {}".format(identity.Path()))

        existing = self.Get(identity)
        if existing.Metadata().Kind() != obj.Metadata().Kind():
            raise Exception(constants.ErrObjectIdentityMismatch)

        if existing.PrimaryKey() != obj.PrimaryKey():
            # log.info("primary key changed from {} to {}".format(
            #     existing.PrimaryKey(), obj.PrimaryKey())
            # )

            if (
                existing.Metadata().Identity().Path()
                != obj.Metadata().Identity().Path()
            ):
                raise Exception(constants.ErrObjectIdentityMismatch)

            # see if there is an object with the new primary key value
            target_identity = store.ObjectIdentity(
                "{}/{}".format(existing.Metadata().Kind().lower(), obj.PrimaryKey())
            )

            target = None
            try:
                target = self.Get(target_identity)
            except Exception as e:
                pass

            if target is not None:
                raise Exception(constants.ErrObjectExists)

        db = self._connection()

        try:
            # Start a transaction
            db.execute("BEGIN")
            cursor = db.cursor()

            self._removeIdentity(cursor, existing.Metadata().Identity().Path())

            obj.Metadata().SetIdentity(existing.Metadata().Identity())

            self._setIdentity(
                cursor,
                obj.Metadata().Identity().Path(),
                obj.PrimaryKey(),
                obj.Metadata().Kind(),
            )

            self._removeObject(
                cursor, existing.PrimaryKey(), existing.Metadata().Kind()
            )

            self._setObject(cursor, obj.PrimaryKey(), obj.Metadata().Kind(), obj)

            db.commit()
            return obj.Clone()
        except Exception as e:
            db.rollback()
            raise e

    def Delete(self, identity, *opt):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)

        log.info("delete {}".format(identity.Path()))
        copt = options.CommonOptionHolderFactory()
        for o in opt:
            o.ApplyFunction()(copt)

        existing = None
        if copt.filter is None:
            existing = self.Get(identity)

        db = self._connection()

        # try:
        # Start a transaction
        db.execute("BEGIN")
        cursor = db.cursor()

        if copt.filter is None:
            self._removeIdentity(cursor, existing.Metadata().Identity().Path())
            self._removeObject(
                cursor, existing.PrimaryKey(), existing.Metadata().Kind()
            )
        else:
            clause = self._buildFilterClause(copt, identity)
            keys = self._getObjectKeys(cursor, identity.Type(), clause)

            self._removeObjects(cursor, identity.Type(), clause)
            self._removeIdentities(cursor, identity.Type(), keys)

        db.commit()
        # except Exception as e:
        #     db.rollback()
        #     raise e

    def Get(self, identity, *opt):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)

        log.info("get {}".format(identity.Path()))

        copt = options.CommonOptionHolderFactory()
        for o in opt:
            o.ApplyFunction()(copt)

        db = self._connection()
        cursor = db.cursor()

        res = None
        if identity.IsId():
            pkey, typ = self._getIdentity(cursor, identity.Path())
            res = self._getObject(cursor, pkey, typ)
        else:
            res = self._getObject(cursor, identity.Key(), identity.Type())
        
        db.commit()
        return res

    def List(self, identity, *opt: list[options.ListOption]):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)

        log.info("list {}".format(identity.Path()))

        if len(identity.Key()) > 0:
            raise Exception(constants.ErrInvalidPath)

        copt = options.CommonOptionHolderFactory()

        for o in opt:
            o.ApplyFunction()(copt)

        db = self._connection()
        cursor = db.cursor()

        query = """SELECT Object FROM Objects 
        WHERE Type = '{}'""".format(
            identity.Type()
        )

        query += self._buildFilterClause(copt, identity)

        if copt.order_by is not None and len(copt.order_by) > 0:
            query += """
            ORDER BY json_extract(Object, '$.{}')""".format(
                copt.order_by
            )
            if copt.order_incremental is None or copt.order_incremental:
                query = query + " ASC"
            else:
                query = query + " DESC"

        if copt.page_size is not None and copt.page_size > 0:
            query = query + " LIMIT {}".format(copt.page_size)

        if copt.page_offset is not None and copt.page_offset > 0:
            query = query + " OFFSET {}".format(copt.page_offset)

        self._do_query(cursor, query)
        rows = cursor.fetchall()

        res = self._parseObjectRows(rows, identity.Type())
        db.commit()
        return res

    def _prepareTables(self, db):
        create = """
        CREATE TABLE IF NOT EXISTS IdIndex (
            Path VARCHAR(75) NOT NULL PRIMARY KEY,
            Pkey NVARCHAR(50) NOT NULL,
            Type VARCHAR(25) NOT NULL);
        """

        cursor = db.cursor()
        self._do_query(cursor, create)

        create = """
        CREATE TABLE IF NOT EXISTS Objects (
            Pkey NVARCHAR(50) NOT NULL,
            Type VARCHAR(25) NOT NULL,
            Object JSON,
            PRIMARY KEY (Pkey,Type));
        """

        cursor.execute(create)
        cursor.close()

        db.commit()

    def _getIdentity(self, cursor, path):
        query = """SELECT Pkey, Type FROM IdIndex 
        WHERE Path='{}'""".format(
            path
        )

        self._do_query(cursor, query)
        result = cursor.fetchone()

        if result is not None:
            pkey, typ = result
            return pkey, typ
        else:
            raise Exception(constants.ErrNoSuchObject)

    def _setIdentity(self, cursor, path, pkey, typ):
        # existing_pkey, existing_typ = None, None

        # try:
        #     existing_pkey, existing_typ = self._getIdentity(cursor, path)
        # except Exception as e:
        #     log.debug("identity get failed: {}".format(e))

        # query = ""
        # if existing_pkey is not None and existing_typ is not None:
        #     query = """UPDATE IdIndex SET Pkey='{}', Type='{}'
        #     WHERE Path='{}'""".format(
        #         pkey, typ.lower(), path
        #     )
        # else:
        query = """INSERT INTO IdIndex (Pkey, Type, Path)
        VALUES ('{}', '{}', '{}')""".format(
            pkey, typ.lower(), path
        )

        self._do_query(cursor, query)

    def _removeIdentity(self, cursor, path):
        query = """DELETE FROM IdIndex
        WHERE Path = '{}'""".format(
            path
        )

        self._do_query(cursor, query)

    def _getObject(self, cursor, pkey, typ):
        query = """SELECT Object FROM Objects
        WHERE Pkey='{}' AND Type='{}'""".format(
            pkey, typ.lower()
        )

        self._do_query(cursor, query)
        result = cursor.fetchone()

        if result is not None:
            data = result[0]
            data = utils.decode_string(data)
            return self._parseObjectRow(data, typ)
        else:
            raise Exception(constants.ErrNoSuchObject)

    def _setObject(self, cursor, pkey, typ, obj):
        query = ""
        # existing_obj = None

        # try:
        #     existing_obj = self._getObject(cursor, pkey, typ)
        # except Exception as e:
        #     log.debug("object get failed: {}".format(e))

        data = obj.ToJson()
        # encode the data so it doesn't contain any SQL unfriendly characters
        data = utils.encode_string(data)

        # if existing_obj is not None:
        #     query = """UPDATE Objects SET Object='{}'
        #     WHERE Pkey = '{}' AND Type = '{}'""".format(
        #         data, pkey, typ.lower()
        #     )
        # else:
        query = """INSERT INTO Objects (Object, Pkey, Type)
        VALUES ('{}', '{}', '{}')""".format(
            data, pkey, typ.lower()
        )

        self._do_query(cursor, query)

    def _removeObject(self, cursor, pkey, typ):
        query = """DELETE FROM Objects
        WHERE Pkey = '{}' AND Type = '{}'""".format(
            pkey, typ.lower()
        )

        self._do_query(cursor, query)

    def _getObjectKeys(self, cursor, typ, clause):
        query = """SELECT Pkey FROM Objects
        WHERE Type = '{}' {}""".format(
            typ.lower(),
            clause,
        )

        self._do_query(cursor, query)
        rows = cursor.fetchall()
        res = []
        for row in rows:
            res.append(row[0])
        return res

    def _removeIdentities(self, cursor, typ, keys):
        batch_size = 100
        for i in range(0, len(keys), batch_size):
            batch = keys[i : i + batch_size]
            clause = "Pkey IN ({})".format(",".join(["'{}'".format(k) for k in batch]))

            query = """DELETE FROM IdIndex
                WHERE Type = '{}' AND {}""".format(
                typ.lower(),
                clause,
            )

            self._do_query(cursor, query)

    def _removeObjects(self, cursor, typ, clause):
        query = """DELETE FROM Objects
        WHERE Type = '{}' {}""".format(
            typ.lower(),
            clause,
        )

        self._do_query(cursor, query)

    def _parseObjectRow(self, data, typ):
        return utils.unmarshal_object(data, self._schema, typ)

    def _parseObjectRows(self, rows, typ):
        res = []
        for row in rows:
            data = utils.decode_string(row[0])
            res.append(self._parseObjectRow(data, typ))
        return res

    def _do_query(self, cursor, query):
        log.debug("running query: {}".format(query))

        statements = sqlparse.parse(query)
        if len(statements) > 1:
            raise Exception(constants.ErrInvalidRequest)

        cursor.execute(query)

    def _buildFilterClause(self, copt, identity):
        if copt.filter is None:
            return ""

        sample = self._schema.ObjectForKind(identity.Type())
        # if sample is None:
        #     raise Exception(constants.ErrInvalidPath)

        return """
            AND ({})
            """.format(
            self._convertFilter(copt.filter, sample)
        )

    def _convertFilter(self, filterOption, sample):
        if isinstance(filterOption, options._ListDeleteOption):
            copt = options.CommonOptionHolderFactory()
            filterOption.ApplyFunction()(copt)
            filterOption = copt.filter

        if isinstance(filterOption, options.AndOption):
            return "( {} )".format(
                " AND ".join(
                    [self._convertFilter(f, sample) for f in filterOption.filters]
                )
            )

        if isinstance(filterOption, options.OrOption):
            return "( {} )".format(
                " OR ".join(
                    [self._convertFilter(f, sample) for f in filterOption.filters]
                )
            )

        if isinstance(filterOption, options.NotOption):
            return "( NOT {} )".format(
                self._convertFilter(filterOption.filter, sample)
            )

        if utils.object_path(sample, filterOption.key) is None:
            raise Exception(constants.ErrInvalidFilter)

        def convert_value(v):
            # return "'{}'".format(v)
            if isinstance(v, str):
                return "'{}'".format(utils.encode_string(v))
            elif isinstance(v, bool):
                return str(v).lower()
            else:
                return str(v)
        
        if isinstance(filterOption, options.InOption):
            values = ", ".join([convert_value(v) for v in filterOption.values])

            return " json_extract(Object, '$.{}') IN ({})".format(
                filterOption.key, values
            )

        if isinstance(filterOption, options.EqOption):
            return " json_extract(Object, '$.{}') = {} ".format(
                filterOption.key, convert_value(filterOption.value)
            )

        if isinstance(filterOption, options.LtOption):
            return " json_extract(Object, '$.{}') < {} ".format(
                filterOption.key, convert_value(filterOption.value)
            )

        if isinstance(filterOption, options.GtOption):
            return " json_extract(Object, '$.{}') > {} ".format(
                filterOption.key, convert_value(filterOption.value)
            )

        if isinstance(filterOption, options.LteOption):
            return " json_extract(Object, '$.{}') <= {} ".format(
                filterOption.key, convert_value(filterOption.value)
            )

        if isinstance(filterOption, options.GteOption):
            return " json_extract(Object, '$.{}') >= {} ".format(
                filterOption.key, convert_value(filterOption.value)
            )

        raise Exception(constants.ErrInvalidFilter)



import os
import re
import logging

log = logging.getLogger(__name__)
log.debug('loading utils.py...')


def yaml_files(path):
    res = []

    lib_regex = re.compile(r"^\w+\.(?:yaml|yml)$")

    for root, dirs, files in os.walk(path):
        for name in files:
            if lib_regex.match(name):
                log.debug(f"found yaml file: {name}")
                res.append(os.path.join(root, name))

    return res

def capitalize(s):
    return s[:1].upper() + s[1:]

def decapitalize(s):
    return s[:1].lower() + s[1:]

def read_file(path):
    with open(path, "rb") as f:
        return f.read()



import os
import shutil
import black
import logging
from io import StringIO

from jinja2 import Template

from pystorz.mgen.loader import load_model, Resource, Struct, Prop, typeDefault
from pystorz.store import utils

log = logging.getLogger(__name__)
log.debug('loading builder.py...')


def Generate(*models) -> None:
    structs = []
    resources = []

    for model in models:
        log.debug(f"Loading model {model}...")
        
        s, r = load_model(model)
        
        structs.extend(s)
        resources.extend(r)

    imports = [
        "import json",
        "from pystorz.store import utils",
        "from pystorz.store import store",
        "from pystorz.store import meta",
    ]

    b = StringIO()

    b.write(render("mgen/templates/imports.py", {"imports": imports}))
    b.write(compileStructs(structs))
    b.write(compileResources(resources))

    res = b.getvalue().replace("&#34;", "\"")

    # refactor and format python code
    res = reformat_python_code(res)

    targetDir = "generated"

    if os.path.exists(targetDir):
        # delete the directory with contents
        shutil.rmtree(targetDir)

    utils.export_file(targetDir, "model.py", res)


def compileResources(resources: list[Resource]) -> str:
    b = StringIO()

    for r in resources:
        log.debug(f"Compiling resource {r.name}...")

        props = [
            Prop(
                "Meta",
                "store.Meta",
                "metadata",
                f'meta.MetaFactory("{r.name}")')
        ]

        if r.external:
            props.append(Prop(
                "External",
                r.external,
                "external"))

        if r.internal:
            props.append(Prop(
                "Internal",
                r.internal,
                "internal"))

        s = Struct(
            r.name,
            r.name,
            props,
        )

        b.write(compileStruct(s))
        b.write(render("mgen/templates/meta.py", r))

    b.write(render("mgen/templates/schema.py", {"resources": resources}))

    return b.getvalue()


def compileStructs(structs: list[Struct]) -> str:
    b = StringIO()

    for s in structs:
        b.write(compileStruct(s))

    return b.getvalue()


def compileStruct(s: Struct) -> str:
    log.debug(f"Compiling struct {s.name}...")

    b = StringIO()
    methods = []

    s.properties = addDefaultPropValues(s.properties)

    parent = "" # s.implements

    for p in s.properties:
        if p.name != "Meta":
            # methods.append(f"{p.name}(self) -> {p.StrippedType()}")
            methods.append(f"{p.name}(self)")

        if p.name != "Meta" and p.name != "External" and p.name != "Internal":
            # methods.append(f"Set{p.name}(self, val: {p.StrippedType()})")
            methods.append(f"Set{p.name}(self, val)")

        if p.name == "External":
            parent = "store.ExternalHolder"
    
    b.write(
        render(
            "mgen/templates/interface.py",
            {
                'name': s.name,
                'methods': methods,
                'implements': parent
            }))

    for p in s.properties:
        if p.name == "External":
            b.write(render("mgen/templates/specinternal.py", None))

    b.write(render("mgen/templates/structure.py", s))
    b.write(render("mgen/templates/unmarshall.py", s))

    return b.getvalue()


def render(rpath: str, data) -> str:
    path = os.path.join(utils.runtime_dir(), rpath)

    with open(path, 'r') as f:
        content = f.read()

    t = Template(content)
    return t.render(data=data)


def addDefaultPropValues(props: list[Prop]) -> list[Prop]:
    res = []

    for p in props:
        if len(p.default) > 0:
            res.append(p)
            continue

        res.append(Prop(p.name, p.type, p.json, typeDefault(p.type)))

    return res


def reformat_python_code(code_str):
    try:
        # Reformat the code string using the parsed AST.
        formatted_code = black.format_str(code_str, mode=black.FileMode())

        return formatted_code

    except Exception as e:
        log.error("failed to reformat code:", e)
        return code_str



 

def {{ data.name }}Factory():
	ret = _{{ data.name }}()
	{% for prop in data.properties %}
	ret.{{ prop.name }}_ = {{ prop.default }}{% endfor %}
	
	return ret


class _{{ data.name }}({{data.implements}}):

	def __init__(self):
		{% for prop in data.properties %}
		self.{{ prop.name }}_ = {{prop.default}}{% endfor %}

	{% for prop in data.properties %}
	{% if prop.name != "External" %}

	def Set{{ prop.name }}(self, val):
		{% if prop.type == "datetime" %}
		self.{{ prop.name }}_ = utils.datetime_string(val)
		{% elif prop.type == "string" %}
		self.{{ prop.name }}_ = str(val)
		{% elif prop.type == "int" %}
		self.{{ prop.name }}_ = int(val)
		{% elif prop.type == "float" %}
		self.{{ prop.name }}_ = float(val)
		{% elif prop.type == "bool" %}
		self.{{ prop.name }}_ = bool(val)
		{% else %}
		self.{{ prop.name }}_ = val
		{% endif %}

	def {{ prop.name }}(self):
		{% if prop.type == "datetime" %}
		return utils.datetime_parse(self.{{ prop.name }}_)
		{% else %}
		return self.{{ prop.name }}_
		{% endif %}

	{% endif %}
	{% endfor %}

	def Clone(entity):
		return utils.clone_object(entity, Schema())




import json
import json
import uuid
import logging
import requests

from pystorz.rest import server, headers
from pystorz.store import store, utils, options
from pystorz.internal import constants

# from urllib.parse import urljoin, urlunparse, urlencode
from urllib.parse import quote, urlparse


log = logging.getLogger(__name__)


class RestOptions:
    def __init__(self):
        self.common_option_holder = options.CommonOptionHolderFactory()
        self.headers = {}

    def common_options(self):
        return self.common_option_holder


def new_rest_options(d):
    res = RestOptions()
    res.headers = {}

    for h in d.headers:
        h.ApplyFunction()(res)

    return res


# def make_http_request(path, content, request_type, headers):
#     requests.packages.urllib3.disable_warnings()  # Disable SSL verification warning

#     req = requests.Request(request_type, path, data=content)

#     for k, v in headers.items():
#         req.headers[k] = v

#     with requests.Session() as session:
#         resp = session.send(req.prepare(), verify=False)

#         rd = resp.content

#         if resp.status_code < 200 or resp.status_code >= 300:
#             return rd, f"http {resp.status_code}"

#         return rd, None


def error_check(response):
    if len(response) == 0:
        return

    data = json.loads(response)
    if "errors" in data:
        raise Exception(data["errors"][0])

    if "error" in data:
        raise Exception(data["error"])


def make_path_for_type(base_url, obj):
    return "{}/{}".format(base_url, obj.Metadata().Kind().lower())


def remove_trailing_slash(val):
    if val.endswith("/"):
        return val[:-1]
    return val


def make_path_for_identity(base_url, identity, params):
    if len(params) > 0:
        return f"{base_url}/{remove_trailing_slash(identity.Path().lower())}?{params}"

    return f"{base_url}/{identity.Path().lower()}"


def list_parameters(ropt):
    opt = ropt.common_options()

    q = {}
    if opt.order_by:
        q[server.OrderByArg] = opt.order_by
        q[server.IncrementalArg] = str(opt.order_incremental).lower()

    if opt.page_offset and opt.page_offset > 0:
        q[server.PageOffsetArg] = str(opt.page_offset)

    if opt.page_size and opt.page_size > 0:
        q[server.PageSizeArg] = str(opt.page_size)

    # if opt.filter:
    #     q[server.FilterArg] = opt.filter.ToJson()

    return "&".join([f"{k}={quote(v)}" for k, v in q.items()])


def filter_parameter(ropt):
    opt = ropt.common_options()
    if opt.filter:
        return opt.filter.ToJson()

    return ""


def strip_serialize(object):
    data = object.ToDict()
    res = {}
    if "external" in data:
        res["external"] = data["external"]
    if "metadata" in data:
        res["metadata"] = data["metadata"]
    
    return json.dumps(res)


class Client(store.Store):
    def __init__(self, base_url, schema, *header_options: list[headers._HeaderOption]):
        self.base_url = base_url
        self.schema = schema
        self.headers = header_options

    def Create(self, obj, *opt):
        if obj is None:
            raise ValueError(constants.ErrObjectNil)

        log.info("get {}".format(obj.Metadata().Identity().Path()))

        copt = new_rest_options(self)
        for o in opt:
            o.ApplyFunction()(copt)

        data = self._process_request(
            make_path_for_type(self.base_url, obj),
            strip_serialize(obj),
            server.ActionCreate,
            copt.headers,
        )

        clone = obj.Clone()
        clone.FromJson(data)

        return clone

    def Get(self, identity, *opt):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)

        log.info("get {}".format(identity.Path()))

        copt = new_rest_options(self)
        for o in opt:
            o.ApplyFunction()(copt)

        resp = self._process_request(
            make_path_for_identity(self.base_url, identity, ""),
            b"",
            server.ActionGet,
            copt.headers,
        )

        tp = identity.Type()
        if tp == "id":
            tp = utils.object_kind(json.loads(resp))

        return utils.unmarshal_object(resp, self.schema, tp)

    def Update(self, identity, obj, *opt):
        if obj is None:
            raise Exception(constants.ErrObjectNil)

        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        
        log.info("update {}".format(identity.Path()))

        copt = new_rest_options(self)
        for o in opt:
            o.ApplyFunction()(copt)
        
        # if identity.Type() != obj.Metadata().Kind():
        #     log.debug("identity type: {}, object type: {}".format(
        #         identity.Type(),
        #         obj.Metadata().Kind()))
        #     raise Exception(constants.ErrObjectIdentityMismatch)

        data = self._process_request(
            make_path_for_identity(self.base_url, identity, ""),
            strip_serialize(obj),
            server.ActionUpdate,
            copt.headers,
        )

        clone = obj.Clone()
        clone.FromJson(data)

        return clone

    def Delete(self, identity, *opt):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        
        log.info("delete {}".format(identity.Path()))

        copt = new_rest_options(self)
        for o in opt:
            o.ApplyFunction()(copt)

        params = list_parameters(copt)
        body = filter_parameter(copt)
        path = make_path_for_identity(self.base_url, identity, params)

        res = self._process_request(
            path,
            body,
            server.ActionDelete,
            copt.headers,
        )

        if res and len(res) > 0:
            log.debug("delete response: {}".format(res))

        return None

    def List(self, identity, *opt):
        if identity is None:
            raise Exception(constants.ErrInvalidPath)

        if len(identity.Key()) > 0:
            raise Exception(constants.ErrInvalidPath)

        log.info("list {}".format(identity.Path()))

        copt = new_rest_options(self)
        for o in opt:
            o.ApplyFunction()(copt)

        params = list_parameters(copt)
        body = filter_parameter(copt)

        path = make_path_for_identity(self.base_url, identity, params)
        res = self._process_request(path, body, server.ActionGet, copt.headers)

        parsed = json.loads(res)

        marshalledResult = store.ObjectList()
        if len(parsed) == 0:
            return marshalledResult

        resource = self.schema.ObjectForKind(utils.object_kind(parsed[0]))

        for r in parsed:
            clone = resource.Clone()
            clone.FromDict(r)

            marshalledResult.append(clone)

        return marshalledResult

    def _make_request(self, path, content, request_type, headers):
        # headers.update(self.headers)
        log.debug(f"making {request_type} request to {path}")

        response = requests.request(
            request_type, path, data=content, headers=headers
        )

        # response.raise_for_status()

        return response.content

    def _process_request(self, request_url, content, method, headers):
        req_id = str(uuid.uuid4())

        request_url = request_url.lower()
        url = urlparse(request_url)
        path = url.path
        new_path = path.replace("//", "/")
        request_url = request_url.replace(path, new_path)

        headers["Origin"] = request_url.replace(url.path, "").replace(url.query, "")
        headers["X-Request-ID"] = req_id
        headers["Content-Type"] = "application/json"
        headers["X-Requested-With"] = "XMLHttpRequest"

        data = self._make_request(request_url, content, method, headers)
        try:
            error_check(data)
        except Exception as err:
            if len(content) > 0:
                try:
                    js = json.loads(content)
                    log.info("request content: %s", json.dumps(js))
                except ValueError:
                    log.info("request content: %s", content)

            if len(data) > 0:
                log.info("response content: %s", data)

            raise err

        return data



import uuid
import json


class Object:

    def __init__(self):
        raise Exception("Object is an interface")

    def Metadata(self):
        pass

    def Clone(self):
        pass

    def ToJson(self) -> str:
        pass

    def FromJson(self, jstr):
        pass

    def FromDict(self, dict) -> Exception:
        pass

    def ToDict(self) -> dict:
        pass

    def PrimaryKey(self) -> str:
        pass


class ExternalHolder(Object):
    def ExternalInternalSet(self, obj: object):
        pass

    def ExternalInternal(self) -> object:
        pass


class ObjectList(list[Object]):
    pass


class ObjectIdentity:
    def __init__(self, id: str):
        self.id_ = id

    def FromString(self, str) -> Exception:
        self.id_ = str

    def __str__(self) -> str:
        return self.id_

    def Path(self) -> str:
        if '/' in self.id_:
            tokens = self.id_.split('/')
            return f"{tokens[0].lower()}/{tokens[1]}"
        else:
            return f"id/{self}"

    def IsId(self) -> bool:
        return '/' not in self.id_

    def Type(self) -> str:
        return self.Path().split('/')[0]

    def Key(self) -> str:
        tokens = self.Path().split('/')
        if len(tokens) > 1:
            return tokens[1]
        else:
            return ""

    def __eq__(self, other):
        if isinstance(other, ObjectIdentity):
            return self.id_ == other.id_
        elif isinstance(other, str):
            return self.id_ == other
        
        return False

    def __hash__(self):
        return hash(self.id_)
    

    def __len__(self):
        return len(self.id_)


def ObjectIdentityFactory() -> ObjectIdentity:
    id = str(uuid.uuid1())
    id = id.replace("-", "")
    
    return ObjectIdentity(id)


class Store:
    def Get(self, identity: ObjectIdentity, *options) -> Object:
        pass

    def List(self, identity: ObjectIdentity, *options) -> ObjectList:
        pass

    def Create(self, obj: Object, *options) -> Object:
        pass

    def Delete(self, identity: ObjectIdentity, *options) -> Exception:
        pass

    def Update(self, identity: ObjectIdentity, obj: Object, *options) -> Object:
        pass


class SchemaHolder:
    def ObjectForKind(self, kind: str) -> Object:
        pass

    def Types(self) -> list[str]:
        pass


class Factory:
    def __call__(self, schema: SchemaHolder) -> Store:
        pass


def New(schema: SchemaHolder, factory: Factory) -> Store:
    return factory(schema)


class StoreJsonDecoder(json.JSONDecoder):

    def __init__(self, schema: SchemaHolder):
        super().__init__(object_hook=self.object_hook)
        self.schema = schema

    def object_hook(self, dct):
        # find the class
        class_name = dct["meta"]["kind"]
        instance = self.schema.ObjectForKind(class_name)
        if instance is None:
            raise Exception(f"cannot find kind: {class_name}")

        return instance.FromJson(dct)




class {{ data["name"] }}({{ data["implements"] }}):

	def __init__(self):
		raise Exception("cannot initialize like this. use the factory method")

{% for method in data["methods"] %}
	def {{ method }}:
		pass

{% endfor %}




ErrObjectNil = "object is nil"
ErrInvalidMethod = "method not allowed"
ErrObjectExists = "object already exists"
ErrNoSuchObject = "object does not exist"
ErrObjectIdentityMismatch = "object identity mismatch"
ErrInvalidFilter = "invalid filter key"
ErrInvalidPath = "invalid request path"
ErrInvalidRequest = "invalid request content"

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"



import json
import yaml
import logging

from pystorz.mgen.utils import capitalize, decapitalize, yaml_files

log = logging.getLogger(__name__)
log.debug("loading loader.py...")


def jopp(obj):
    # json pretty print
    jstring = """JSON pretty print
""" + json.dumps(
        obj, indent=4, sort_keys=True
    )
    return jstring


class Model:
    def __init__(self, types=None):
        self.types = types or []


class Prop:
    def __init__(self, name, prop_type, json_str, default=""):
        self.name = name
        self.type = prop_type
        self.json = json_str
        self.default = default

    def IsMap(self):
        if len(self.type) < 3:
            return False

        return self.type[:3] == "map"

    def IsArray(self):
        if len(self.type) < 2:
            return False

        return self.type[:2] == "[]"

    def StrippedType(self):
        if self.IsMap():
            return "dict"
        if self.IsArray():
            return "list"
        if self.type == "string":
            return "str"
        return self.type

    def ComplexTypeValueDefault(self):
        return complexTypeValueDefault(self.type)

    def StrippedDefault(self):
        return typeDefault(self.StrippedType())


class Struct:
    def __init__(self, name, implements="", properties=[]):
        self.name = name
        self.implements = implements
        self.properties = properties


class Resource:
    def __init__(self, name, external=None, internal=None, primary_key=None):
        self.name = name
        self.external = external
        self.internal = internal
        self.primary_key = primary_key

    def IdentityPrefix(self):
        return self.name.lower()


def load_model(path: str):
    yamls = yaml_files(path)
    structs = []
    resources = []
    struct_cache = set()
    resource_cache = set()

    for y in yamls:
        model = read_model(y)

        # log.debug("model: {}".format(jopp(model)))

        for m in model["types"]:
            if m["kind"].lower() == "struct":
                if m["name"] in struct_cache:
                    raise Exception("duplicate struct: {}".format(m["name"]))
                struct_cache.add(m["name"])

                structs.append(Struct(m["name"], "", capitalize_props(m["properties"])))
                continue
            if m["kind"].lower() == "object":
                if m["name"] in resource_cache:
                    raise Exception("duplicate object: {}".format(m["name"]))
                resource_cache.add(m["name"])

                pkey = "metadata.identity"
                if "primarykey" in m:
                    pkey = m["primarykey"]
                pkey = make_prop_caller_string(pkey)

                ext = None
                if "external" in m:
                    ext = m["external"]
                intr = None
                if "internal" in m:
                    intr = m["internal"]

                resources.append(Resource(m["name"], ext, intr, pkey))
                continue

            raise Exception("unknown kind: {}".format(m["kind"]))

    errors = validate_model(structs, resources)
    if len(errors) > 0:
        raise Exception(
            """model validation failed:
    {}""".format(
                """
    """.join(
                    errors
                )
            )
        )

    return structs, resources


def validate_model(structs, resources):
    errors = []
    known_types = ["string", "int", "float", "bool", "datetime"]

    for s in structs:
        known_types.append(s.name)

    for s in structs:
        for p in s.properties:
            if p.IsArray():
                elem_type = p.type[2:]
                if elem_type not in known_types:
                    errors.append(
                        "struct {} property {}: unknown type: {}".format(
                            s.name, p.name, elem_type
                        )
                    )
                continue

            if p.IsMap():
                # map[int]string
                key_type = p.type[4:].split("]")[0]
                val_type = p.type[4:].split("]")[1]
                if key_type not in known_types:
                    errors.append(
                        "struct {} property {}: unknown type: {}".format(
                            s.name, p.name, elem_type
                        )
                    )
                if val_type not in known_types:
                    errors.append(
                        "struct {} property {}: unknown type: {}".format(
                            s.name, p.name, elem_type
                        )
                    )
                continue

            if p.type not in known_types:
                errors.append(
                    "struct {} property {}: unknown type: {}".format(
                        s.name, p.name, p.type
                    )
                )

    for r in resources:
        if r.external is None and r.internal is None:
            errors.append("resource {} has no internal and external".format(r.name))
            continue

        if r.external is not None:
            if r.external not in known_types:
                errors.append(
                    "resource {} external: unknown type: {}".format(r.name, r.external)
                )

        if r.internal is not None:
            if r.internal not in known_types:
                errors.append(
                    "resource {} internal: unknown type: {}".format(r.name, r.internal)
                )

    return errors


def read_model(path: str):
    log.debug(f"reading model from {path}")

    with open(path, "r") as f:
        data = yaml.safe_load(f)
    return data


def capitalize_props(l: list):
    res = []
    for p in l:
        res.append(
            Prop(
                capitalize(p["name"]),
                p["type"],
                decapitalize(p["name"]),
                # p.default
            )
        )
    return res


def make_prop_caller_string(pkey: str):
    tok = pkey.split(".")
    cap = []
    for t in tok:
        cap.append("{}()".format(capitalize(t)))

    return ".".join(cap)


def typeDefault(tp: str) -> str:
    if tp.startswith("list") or tp.startswith("[]"):
        return "list()"
        # return f"{tp} {{}}"
    if tp.startswith("dict") or tp.startswith("map"):
        return "dict()"
        # return f"make({tp})"

    if tp == "str" or tp == "string":
        return '""'
    if tp == "bool":
        return "False"
    if tp == "int":
        return "0"
    if tp == "float":
        return "0.0"
    if tp == "datetime":
        return '"0001-01-01T00:00:00.000000Z"'

    return f"{tp}Factory()"


def complexTypeValueDefault(tp: str) -> str:
    log.debug(f"typeValueDefault: {tp}")

    if tp.startswith("[]"):
        return typeDefault(tp[2:])

    if tp.startswith("map"):
        closing_bracket_index = tp.find("]")
        if closing_bracket_index == -1:
            raise Exception("invalid map type: {}".format(tp))

        return typeDefault(tp[closing_bracket_index + 1 :])

    raise Exception("unknown type: {}".format(tp))



import logging
from pystorz.store import store
from pystorz.store import options
from pystorz.internal import constants

log = logging.getLogger(__name__)


class RouteStore(store.Store):
    def __init__(self, default: store.Store, mapping: dict):
        self.Default = default
        self.Mapping = {}

        for k, v in mapping.items():
            self.Mapping[k.lower().replace("/", "")] = v

    def Create(self, obj: store.Object, *opt: options.CreateOption) -> store.Object:
        if obj is None:
            raise Exception(constants.ErrObjectNil)

        log.info(f"create {obj.PrimaryKey()}")

        return self._getStore(obj.Metadata().Kind()).Create(obj, *opt)

    def Update(self, identity: store.ObjectIdentity, obj: store.Object, *opt: options.UpdateOption) -> store.Object:
        if obj is None:
            raise Exception(constants.ErrObjectNil)

        if identity is None:
            raise Exception(constants.ErrInvalidPath)

        log.info(f"update {identity.Path()}")

        return self._getStore(obj.Metadata().Kind()).Update(identity, obj, *opt)

    def Delete(self, identity: store.ObjectIdentity, *opt: options.DeleteOption) -> None:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        
        log.info(f"delete {identity.Path()}")

        return self._getStore(identity.Type()).Delete(identity, *opt)

    def Get(self, identity: store.ObjectIdentity, *opt: options.GetOption) -> store.Object:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)
        
        log.info(f"get {identity.Path()}")

        return self._getStore(identity.Type()).Get(identity, *opt)

    def List(self, identity: store.ObjectIdentity, *opt: options.ListOption) -> store.ObjectList:
        if identity is None:
            raise Exception(constants.ErrInvalidPath)

        log.info(f"list {identity.Type()}")

        return self._getStore(identity.Type()).List(identity, *opt)

    def _getStore(self, kind: str) -> store.Store:
        kind = kind.lower().replace("/", "")

        if kind in self.Mapping:
            return self.Mapping[kind]

        if self.Default is None:
            raise Exception(constants.ErrInvalidPath)

        log.debug(f"kind {kind} not found in mapping, using default store")
        return self.Default



import logging
import mysql.connector
from pystorz.sql.store import SqlStore

log = logging.getLogger(__name__)

class _MySQLAdapter:
    def __init__(self, host, port, username, password, database):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.connection = None
        self._cursor = None

    def connect(self):
        try:
            self._cursor = None
            self.connection = mysql.connector.connect(
                host=self.host,
                port=self.port,
                user=self.username,
                password=self.password,
                database=self.database
            )
            return self.connection
        except Exception as err:
            log.error("connect", err)
            raise err

    def cursor(self):
        try:
            if self.connection is None:
                self.connect()
            
            if self._cursor is None:
                self._cursor = self.connection.cursor()

            return self._cursor
        except Exception as err:
            log.error("cursor", err)
            raise err

    def close(self):
        try:
            if self._cursor:
                self._cursor.close()
            if self.connection:
                self.connection.close()
        except Exception as err:
            log.error("close", err)
            raise err
        self._cursor = None

    def execute(self, query, params=None):
        try:
            if self._cursor is None:
                self.cursor()
            
            if params:
                self._cursor.execute(query, params)
            else:
                self._cursor.execute(query)
        except Exception as err:
            log.error("execute", err)
            raise err

    def fetchall(self):
        try:
            return self._cursor.fetchall()
        except Exception as err:
            log.error("fetch all", err)
            raise err

    def commit(self):
        try:
            self.connection.commit()
        except Exception as err:
            log.error("commit", err)
            raise err
        self._cursor = None

    def rollback(self):
        try:
            self.connection.rollback()
        except Exception as err:
            log.error("rollback", err)
            raise err
        self._cursor = None


def MySqlStoreFactory(schema, host, port, user, password, database):
    def connector():
        return _MySQLAdapter(host, port, user, password, database)

    return SqlStore(schema, connector)
